#pragma once

#include "GameForms.h"
#include "Utilities.h"

class TESRegionManager;
class BSFile;

// 10
class BoundObjectListHead
{
public:
	BoundObjectListHead();
	~BoundObjectListHead();

	UInt32			boundObjectCount;	// 0
	TESBoundObject	* first;			// 4
	TESBoundObject	* last;				// 8
	UInt32			unkC;				// C
};

typedef NiTArray <TESRegion> TESRegionList;

// 428
struct ModInfo {
	UInt32								unk000[2];			// 000
	UInt32 /*NiTPointerMap<TESFile*>*/	* pointerMap;		// 008
	UInt32								unk00C;				// 00C
	BSFile*								unkFile;			// 010
	UInt32								unk014[3];			// 014
	char								name[64];			// 020
	UInt32	unk060[(0x428-0x060) >> 2];						// 060
	bool IsLoaded() const { return true; }
};
STATIC_ASSERT(offsetof(ModInfo, name) == 0x20);
STATIC_ASSERT(offsetof(ModInfo, unk060) == 0x60);
STATIC_ASSERT(sizeof(ModInfo) == 0x428);

struct ModList
{
	tList<ModInfo>		modInfoList;
	UInt32				modCount;
	ModInfo*			modInfoArray[0xFF];
};
STATIC_ASSERT(sizeof(ModList) == 0x408);

#if RUNTIME
#if FALLOUT_VERSION == FALLOUT_VERSION_1_0_15
	const UInt32 _DataHandler_AddForm = 0x00443CC0;
#elif FALLOUT_VERSION == FALLOUT_VERSION_1_1_35
	const UInt32 _DataHandler_AddForm = 0x00444AC0;
#else
	#error unsupported runtime version
#endif
#else
	const UInt32 _DataHandler_AddForm = 0;
#endif


// 5B8
class DataHandler
{
public:
	DataHandler();
	~DataHandler();

	BoundObjectListHead				* boundObjectList;		// 000
	tList<TESPackage>				packageList;			// 004
	tList<TESWorldSpace>			worldSpaceList;			// 00C
	tList<TESClimate>				climateList;			// 014
	tList<TESImageSpace>			imageSpaceList;			// 01C
	tList<TESImageSpaceModifier>	imageSpaceModList;		// 024
	tList<TESWeather>				weatherList;			// 02C
	tList<EnchantmentItem>			enchantmentItemList;	// 034
	tList<SpellItem>				spellItemList;			// 03C
	tList<BGSHeadPart>				headPartList;			// 044
	tList<TESHair>					hairList;				// 04C
	tList<TESEyes>					eyeList;				// 054
	tList<TESRace>					raceList;				// 05C
	tList<BGSEncounterZone>			encounterZoneList;		// 066
	tList<TESLandTexture>			landTextureList;		// 06C
	tList<BGSCameraShot>			cameraShotList;			// 074
	tList<TESClass>					classList;				// 07C
	tList<TESFaction>				factionList;			// 084
	tList<Script>					scriptList;				// 08C
	tList<TESSound>					soundList;				// 094
	tList<BGSAcousticSpace>			acousticSpaceList;		// 09C
	tList<BGSRagdoll>				ragdollList;			// 0A4
	tList<TESGlobal>				globalList;				// 0AC
	tList<BGSVoiceType>				voiceTypeList;			// 0B4
	tList<BGSImpactData>			impactDataList;			// 0BC
	tList<BGSImpactDataSet>			impactDataSetList;		// 0C4
	tList<TESTopic>					topicList;				// 0CC
	tList<TESQuest>					questList;				// 0D4
	tList<TESCombatStyle>			combatStyleList;		// 0DC
	tList<TESLoadScreen>			loadScreenList;			// 0E4
	tList<TESWaterForm>				waterFormList;			// 0EC
	tList<TESEffectShader>			effectShaderList;		// 0F4
	tList<BGSProjectile>			projectileList;			// 0FC
	tList<BGSExplosion>				explosionList;			// 104
	tList<BGSRadiationStage>		radiationStageList;		// 10C
	tList<BGSDebris>				debrisList;				// 114
	tList<BGSPerk>					perkList;				// 11C
	tList<BGSBodyPartData>			bodyPartDataList;		// 124
	tList<BGSNote>					noteList;				// 12C
	tList<BGSListForm>				listFormList;			// 134
	tList<BGSMenuIcon>				menuIconList;			// 13C
	tList<TESObjectANIO>			anioList;				// 144
	tList<BGSMessage>				messageList;			// 14C
	tList<BGSLightingTemplate>		lightninghTemplateList;	// 154
	tList<BGSMusicType>				musicTypeList;			// 15C
	TESRegionList					* regionList;			// 164
	NiTArray<TESObjectCELL*>		cellArray;				// 168
	NiTArray<BGSAddonNode*>			addonArray;				// 178
	UInt32	unk188[0x5];									// 188
	ModList							modList;				// 19C
	UInt32							unk5A4;					// 5A4
	UInt32							unk5A8;					// 5A8
	UInt32							unk5AC;					// 5AC
	TESRegionManager				* regionManager;		// 5B0
	UInt32							unk5B4;					// 5B4
	
	static DataHandler* Get();
	const ModInfo* LookupModByName(const char* modName);
	UInt8 GetModIndex(const char* modName);
	UInt8 GetActiveModCount() const;

	MEMBER_FN_PREFIX(DataHandler);
	DEFINE_MEMBER_FN(DoAddForm, UInt32, _DataHandler_AddForm, TESForm * pForm);	// stupid name is because AddForm is redefined in windows header files

	TESQuest* GetQuestByName(const char* questName);
};

STATIC_ASSERT((0x5A8-0x19C) == 0x40C);
STATIC_ASSERT(sizeof(NiTArray<TESObjectCELL*>) == 0x10);
STATIC_ASSERT(offsetof(DataHandler, listFormList) == 0x134);
STATIC_ASSERT(offsetof(DataHandler, regionList) == 0x164);
STATIC_ASSERT(offsetof(DataHandler, modList) == 0x19C);
STATIC_ASSERT(offsetof(DataHandler, regionManager) == 0x5B0);
STATIC_ASSERT(sizeof(DataHandler) == 0x5B8);
